| -----------------------------------------------------------------------------
|
| This is the 'header.s' file which is appended to the front of the
| assembly program output by the compiler.  It contains run-time routines
| for the scheme system and M4 macro definitions which are needed to
| expand the macros in the compiler's output.

| -----------------------------------------------------------------------------
|
| Representation of objects:
|
| An object is represented using a 32 bit value.  When more than 32 bits
| are needed to represent an object, the 32 bits are actually a pointer
| to the object in memory.
|
|                                  28     28
| Fixnums (integers in the range -2   .. 2  -1):
|
|     XXXXXXXXXXXXXXXXXXXXXXXXXXXXX000
|     \------ integer value ------/
|
| Floating point numbers and other constants:
|
|     XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX1
|     \-- encoding of f.p. number --/
|              or constant
|
| Pair (i.e. cons cell):
|                                           16 bits
|                                          +-------+           |
|                                          |       | \         |
|                                          |- - - -|  > cdr    | high mem.
|                                          |       | /         |
|                                          +-------+           |
|     XXXXXXXXXXXXXXXXXXXXXXXXXXXXX100 --> |       | \         V
|                                          |- - - -|  > car
|                                          |       | /
|                                          +-------+
|
| Hunk (i.e. object in mem. other than pair, e.g. vectors, strings, proc.):
|
|                                           16 bits
|                                          +-------+     |
|                                          |field1 |     |
|                                          |- - - -|     | high mem.
|     XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX10 --> |field2 |     |
|                                          +-------+\    |
|                                          |       | |   V
|                                  dword 1 |- - - -| |
|                                          |       | |
|                                          +-------+ |
|                                     .    |   .   | |
|                                     .    |   .   | length (# of double words)
|                                     .    |   .   | |
|                                          +-------+ |
|                                          |       | |
|                                  dword N |- - - -| |
|                                          |       | |
|                                          +-------+/
|
| meaning of field1 and field2:
|
| field1            field2               type of object
|
| 0TTTTTTTXXXXXXXX  XXXXXXXXXXXXXXXX     TT..TT = type, GC information is
|                                           in a table indexed by TT..TT,
|                                           XX.XX = auxiliary information
| 1000000000000000                       forwarded hunk (dword 1 = forw addr.)
| 10DDDDDDDDDDDDDD  first word of proc.  sub-procedure (DD..DD distance in # of
|                                           quad-bytes from parent procedure)
| 110NNNNNNNNNNNNN  first word of proc.  prim proc. (dword M..dword N =
|                                           GC information)
| 111NNNNNNNNNNNNN  JSR opcode           closure (dword 1 = pointer to body,
|                                           dword 2 ..dword N = closed values)

|------------------------------------------------------------------------------
|
| Scheme constants:
|

Cst_nil		= 0xFFFFFFFF	| -1
Cst_false	= 0xFFFFFFFD	| -3
Cst_true	= 0xFFFFFFFB	| -5
Cst_undefined	= 0xFFFFFFF9	| -7
Forwarded_pair	= 0xFFFFFFF7	| -9, if car of pair is this value, pair has
				|   already been forwarded to new space


|------------------------------------------------------------------------------
|
| Constants:
|

Fix_type_set	= 0x01010101	| value ends with 000
FP_type_set	= 0xAAAAAAAA	| value ends with 1
Pair_type_set	= 0x10101010	| value ends with 100
Hunk_type_set	= 0x44444444	| value ends with 10
Forwarded_hunk	= 0x8000	| if field1 of hunk is this value, the hunk
				|   has already been copied to new space

Vector_type	= 0x40		| type field for vectors
Symbol_type	= 0x41		| type field for symbols
String_type	= 0x60		| type field for strings

Sub_proc_type	= 0x8000	| type field for sub-procedures
Primitive_type	= 0xC000	| type field for primitives
Closure_type	= 0xE000	| type field for closures


|------------------------------------------------------------------------------
|
| Macro definitions:
|

define(concatenate,`$1$2$3$4$5$6$7$8$9')
define(the_number,`0')
define(new_number,`define(`the_number',incr(the_number))the_number')
define(new_symbol,`define(`the_symbol',concatenate($1,new_number))the_symbol')

define(change_area,
`define(`current_area',$1)
	$1')

define(begin_main,
`	.lcomm	_reg_at_entry,4*16
	change_area(.text)
	.globl	_entry
_entry:
	moveml	d0-d7/a0-a7,_reg_at_entry
	movl	sp,d0
	andl	#0xfffffffc,d0
	movl	d0,sp')

define(end_main,
`	moveml	_reg_at_entry,d0-d7/a0-a7
	rts')

define(save_regs,
`	moveml	d0-d7/a0-a6,sp@-')

define(restore_regs,
`	moveml	sp@+,d0-d7/a0-a6')

define(print_str,
`	.data2
new_symbol(Str):	.asciz	$1
	current_area
save_regs
	pea	the_symbol
	jbsr	_printf
	addqw	#0x4,sp
restore_regs')

define(print_int,
`	.data2
new_symbol(Str):	.asciz	$1
	current_area
save_regs
	movl	$2,sp@-
	pea	the_symbol
	jbsr	_printf
	addqw	#0x8,sp
restore_regs')

define(print_chr,
`	.data2
new_symbol(Str):	.asciz	$1
	current_area
save_regs
	movb	$2,d0
	extbl	d0
	movl	d0,sp@-
	pea	the_symbol
	jbsr	_printf
	addqw	#0x8,sp
restore_regs')

define(define_primitive,
`define(`the_procedure_name',$1)define(`the_procedure',scm_$1)
	.align	4
	.word	(end_of_procedure_scm_$1-scm_$1)/4+Primitive_type
scm_$1:')

define(end_primitive,
`	.align	4
concatenate(start_of_constants_,the_procedure):
	.long	new_symbol(Sym)
	.long	(concatenate(end_of_procedure_,the_procedure)-concatenate(start_of_constants_,the_procedure))/4-1
concatenate(end_of_procedure_,the_procedure):
symbol_object(the_symbol,new_symbol(Str))
string_object(the_symbol,"the_procedure_name")')

define(procedure_object_begin,
`define(`the_procedure',$1)
	.align	4
	.word	(end_of_procedure_$1-$1)/4+Primitive_type
$1:')

define(procedure_object_constants,
`	.align	4
concatenate(start_of_constants_,the_procedure):')

define(procedure_object_end,
`	.long	(concatenate(end_of_procedure_,the_procedure)-concatenate(start_of_constants_,the_procedure))/4-1
concatenate(end_of_procedure_,the_procedure):')

define(sub_procedure,
`	.align	4
	.word	($1-the_procedure)/4+Sub_proc_type
$1:')

define(pair_object,
`	.data
	.long	$3
$1:	.long	$2
	current_area')

define(vector_object,
`	.align	4
	.word	Vector_type*256
$1:	.word	$2')

define(symbol_object,
`	.align	4
	.word	Symbol_type*256
$1:	.word	1
	.long	$2')

define(string_object,
`	.align	4
	.word	String_type*256
$1:	.word	end_of_$1-$1-2
	.ascii	`$2'
end_of_$1:')

| -----------------------------------------------------------------------------
|
| Interface for SUN:
|

	change_area(.text)

	.globl	_main
_main:
	link	a6,#0
	jbsr	_entry
	unlk	a6
	rts

begin_main

	lea	data_adr,a0	| make sure pairs are on octuple address

	.data
	.skip	4	| 'magical' offset to bring 'data' section to an
data_adr:		| octuple address (may be different on other systems)
	current_area

	movl	a0,d0
	andl	#7,d0
	beq	pairs_are_ok

print_str("Can not execute; pairs are not at an address multiple of 8\012")
print_str("Change magical offset in header file\012")

	bra	exit

pairs_are_ok:

	movl	#Hunk_type_set,d7	| mask to test tags of hunks
	movl	#Pair_type_set,d6	| mask to test tags of pairs
	movl	#Fix_type_set,d5	| mask to test tags of fixnums

	lea	globals,a6	| setup pointer to global area

	lea	heap_bottom,a5	| setup pointer to allocate hunks
	movl	a5,d0
	andl	#0xFFFFFFF8,d0	| make sure address ends with 000
	movl	d0,a5

	lea	heap_top,a4	| setup pointer to allocate pairs
	movl	a4,d0
	andl	#0xFFFFFFF8,d0	| make sure address ends with 000
	movl	d0,a4

	lea	handler,a3	| error handler

	.lcomm	stack_bottom,41000
	.lcomm	stack_top,0

	lea	stack_bottom,a0	| set stack limit
	movl	a0,a6@

	lea	stack_top,a7	| set stack pointer

	jmp	eval_print	| eval program and print result

exit:

end_main

|------------------------------------------------------------------------------
|
| low-level run-time routines:

rest_enter_2:
	subql	#6,a7@		| fix environment pointer for closures
rest_enter_1:
	cmpl	a6@,a7		| check for stack overflow
	bhis	rest_enter_ok
	bra	error_1
rest_enter_ok:

	movl	a6,a2		| transfer arguments from registers to
	movl	d1,a2@-		| argument vector
	movl	d2,a2@-
	movl	d3,a2@-
	movl	d4,a2@-

	tstw	d0		| get number of actual parameters in d0
	bgts	not_special_code
	addqw	#3,d0
not_special_code:

	movw	d0,d1		| save number of actual parameters

	subw	a0,d0		| are there enough actual parameters?
	bpls	enough_args

print_str("error: too few arguments to an n-ary procedure\012")
	jmp	dump

enough_args:

	subqw	#1,d1		| compute location of last argument in
	extl	d1		| argument vector
	asll	#2,d1
	movl	a6,a2
	subl	d1,a2

	moveq	#Cst_nil,d1	| start with the empty list
	subqw	#1,d0		| just enough actual parameters?
	bmis	push_rest_arg

construct_list_loop:

	movl	a2@+,a4@-
	movl	a4,a0
	movl	d1,a4@-
	movl	a0,d1

	dbra	d0,construct_list_loop

push_rest_arg:

	movl	d1,a7@-
	bras	rest_enter_push

rest_enter_loop:
	movl	a2@+,a7@-
rest_enter_push:
	cmpl	a2,a6
	bnes	rest_enter_loop

	jmp	a1@		| return to procedure

handler:
	movl	a7@+,a0		| get pointer to value of global variable
	movw	d0,a7@-
	movl	a0@,d0
	btst	d0,d7		| is it a procedure?
	beqs	error_3
	movl	d0,a1
	tstw	a1@(-2)
	bpls	error_3

	movw	#20217,a0@(-2)	| change trap to a long 'jmp'
	movw	a7@+,d0		| jump to it...
	jmp	a1@

error_1:
	cmpl	a6@,a7
	bhis	wrong_nb_args

stack_overflow:

print_str("error: stack overflow\012")
	jmp	dump

wrong_nb_args:

print_str("error: wrong number of arguments to a user procedure\012")
	jmp	dump

error_2:
	subql	#6,a7@
	cmpl	a6@,a7
	bhis	wrong_nb_args
	bras	stack_overflow

error_3:
print_str("error: application of a non-procedure object\012")
	jmp	dump

error_4:

print_str("error: wrong number of arguments to a primitive procedure\012")
	jmp	dump

| -----------------------------------------------------------------------------
|
| Heap:

	change_area(.bss)

	.skip	8

heap_bottom:

	.skip	8*400000	| reserve space for heap

heap_top:

| -----------------------------------------------------------------------------
|
| Globals:

	change_area(.data1)

	.word	20217
	.long	rest_enter_1	| entering a procedure with a 'rest' parameter
	.word	20217
	.long	rest_enter_2	| entering a closure with a 'rest' parameter
	.word	20217
	.long	error_1		| wrong number of arg or stack overflow
	.word	20217
	.long	error_2		| wrong number of arg or stack overflow (for a
				| closure)
	.word	20217
	.long	error_3		| calling a non-procedure

	.skip	126*4		| space for 126 arguments

globals:
	.long	0
	.word	20217
	.long	scm_trace
	.word	20217
	.long	scm_make_promise
	.word	20217
	.long	scm_memv
	.word	20217
	.long	scm_cons
	.word	20217
	.long	scm_list
	.word	20217
	.long	scm_append
	.word	20217
	.long	scm_list_vector
	.word	20217
	.long	scm_not
	.word	20217
	.long	scm_eq
	.word	20217
	.long	scm_pair
	.word	20217
	.long	scm_cons
	.word	20217
	.long	scm_append
	.word	20217
	.long	scm_length
	.word	20217
	.long	scm_car
	.word	20217
	.long	scm_cdr
	.word	20217
	.long	scm_set_car
	.word	20217
	.long	scm_set_cdr
	.word	20217
	.long	scm_null
	.word	20217
	.long	scm_num_eq
	.word	20217
	.long	scm_num_lt
	.word	20217
	.long	scm_num_gt
	.word	20217
	.long	scm_plus
	.word	20217
	.long	scm_minus
	.word	20217
	.long	scm_times
	.word	20217
	.long	scm_quotient
	.word	20217
	.long	scm_minus_1
	.word	20217
	.long	scm_force
	.word	20217
	.long	scm_write
	.word	20217
	.long	scm_newline
	.word	20217
	.long	scm_list
	.word	20217
	.long	scm_vect
	.word	20217
	.long	scm_list_vector
	.word	20217
	.long	scm_memv
	.word	20217
	.long	scm_assq
	.word	20217
	.long	scm_symbol
	.word	20217
	.long	scm_vector
	.word	20217
	.long	scm_string
	.word	20217
	.long	scm_procedure
	.word	20217
	.long	scm_number
	.word	20217
	.long	scm_char

	| space for 128 other global variables (initially undefined)

	.word	20115,-1,-7,20115,-1,-7,20115,-1,-7,20115,-1,-7
	.word	20115,-1,-7,20115,-1,-7,20115,-1,-7,20115,-1,-7
	.word	20115,-1,-7,20115,-1,-7,20115,-1,-7,20115,-1,-7
	.word	20115,-1,-7,20115,-1,-7,20115,-1,-7,20115,-1,-7
	.word	20115,-1,-7,20115,-1,-7,20115,-1,-7,20115,-1,-7
	.word	20115,-1,-7,20115,-1,-7,20115,-1,-7,20115,-1,-7
	.word	20115,-1,-7,20115,-1,-7,20115,-1,-7,20115,-1,-7
	.word	20115,-1,-7,20115,-1,-7,20115,-1,-7,20115,-1,-7
	.word	20115,-1,-7,20115,-1,-7,20115,-1,-7,20115,-1,-7
	.word	20115,-1,-7,20115,-1,-7,20115,-1,-7,20115,-1,-7
	.word	20115,-1,-7,20115,-1,-7,20115,-1,-7,20115,-1,-7
	.word	20115,-1,-7,20115,-1,-7,20115,-1,-7,20115,-1,-7
	.word	20115,-1,-7,20115,-1,-7,20115,-1,-7,20115,-1,-7
	.word	20115,-1,-7,20115,-1,-7,20115,-1,-7,20115,-1,-7
	.word	20115,-1,-7,20115,-1,-7,20115,-1,-7,20115,-1,-7
	.word	20115,-1,-7,20115,-1,-7,20115,-1,-7,20115,-1,-7

| -----------------------------------------------------------------------------
|
| Primitive procedures:

define_primitive(trace)
print_str("\012Trace: calling ")
	tstw	d0
	jmp	scm_write
end_primitive

define_primitive(append)
	bnes	scm_append2

	movl	d2,a1

	btst	d1,d6		| is arg1 a pair?
	beqs	scm_append1

	movl	d1,a0
	movl	a0@,a4@-
	movl	a4,a1
	movl	d2,a4@-

scm_append_loop:
	movl	a0@-,d1
	btst	d1,d6		| is the tail a pair?
	beqs	scm_append1

	movl	d1,a0
	movl	a0@,a4@-
	movl	a4,a4@(4)
	movl	d2,a4@-

	bras	scm_append_loop

scm_append1:
	cmpl	#Cst_nil,d1
	bnes	scm_append3

	movl	a1,d1
	rts
scm_append2:
	jmp	error_4		| for now signal an error (should be n-ary)
scm_append3:
print_str("error: improper list passed to 'append'\012")
	jmp	dump
end_primitive

define_primitive(length)
	bpls	scm_length2
	moveq	#0,d2
scm_length_loop:
	btst	d1,d6		| is arg1 a pair?
	beqs	scm_length1

	movl	d1,a0
	movl	a0@-,d1
	addql	#8,d2

	bras	scm_length_loop

scm_length1:
	cmpl	#Cst_nil,d1
	bnes	scm_length3

	movl	d2,d1
	rts
scm_length2:
	jmp	error_4
scm_length3:
print_str("error: improper list passed to 'length'\012")
	jmp	dump
end_primitive

define_primitive(not)
	bpls	scm_not2
	addql	#3,d1
	bcss	scm_not1
	moveq	#Cst_false,d1
	rts
scm_not1:
	moveq	#Cst_true,d1
	rts
scm_not2:
	jmp	error_4
end_primitive

define_primitive(list_vector)
	bpls	scm_list_vector3
	movl	d1,sp@-
	pea	scm_list_vector1
	moveq	#-1,d0
	jmp	scm_length
sub_procedure(scm_list_vector1)
	movl	d1,d2
	asrl	#3,d2
	movw	#Vector_type*256,a5@+
	movl	a5,d1
	movw	d2,a5@+
	movl	sp@+,d2
scm_list_vector_loop:
	btst	d2,d6
	beqs	scm_list_vector2

	movl	d2,a0
	movl	a0@,a5@+
	movl	a0@-,d2

	bras	scm_list_vector_loop

scm_list_vector2:
	rts
scm_list_vector3:
	jmp	error_4
end_primitive

define_primitive(eq)
	bnes	scm_eq2
	cmpl	d2,d1
	beqs	scm_eq1
	moveq	#Cst_false,d1
	rts
scm_eq1:
	moveq	#Cst_true,d1
	rts
scm_eq2:
	jmp	error_4
end_primitive

define_primitive(pair)
	bpls	scm_pair2
	btst	d1,d6
	bnes	scm_pair1
	moveq	#Cst_false,d1
	rts
scm_pair1:
	moveq	#Cst_true,d1
	rts
scm_pair2:
	jmp	error_4
end_primitive

define_primitive(cons)
	bnes	scm_cons1
	movl	d1,a4@-
	movl	a4,d1
	movl	d2,a4@-
	rts
scm_cons1:
	jmp	error_4
end_primitive

define_primitive(car)
	bpls	scm_car2
	btst	d1,d6
	beqs	scm_car1
	movl	d1,a0
	movl	a0@,d1
	rts
scm_car1:
print_str("error: non pair argument to 'car': ")
	moveq	#-1,d0
	jbsr	scm_write
print_str("\012")
	jmp	dump
scm_car2:
	jmp	error_4
end_primitive

define_primitive(cdr)
	bpls	scm_cdr2
	btst	d1,d6
	beqs	scm_cdr1
	movl	d1,a0
	movl	a0@-,d1
	rts
scm_cdr1:
print_str("error: non pair argument to 'cdr': ")
	moveq	#-1,d0
	jbsr	scm_write
print_str("\012")
	jmp	dump
scm_cdr2:
	jmp	error_4
end_primitive

define_primitive(set_car)
	bnes	scm_set_car2
	btst	d1,d6
	beqs	scm_set_car1
	movl	d1,a0
	movl	d2,a0@
	rts
scm_set_car1:
print_str("error: non pair argument to 'set-car': ")
	moveq	#-1,d0
	jbsr	scm_write
print_str("\012")
	jmp	dump
scm_set_car2:
	jmp	error_4
end_primitive

define_primitive(set_cdr)
	bnes	scm_set_cdr2
	btst	d1,d6
	beqs	scm_set_cdr1
	movl	d1,a0
	movl	d2,a0@-
	rts
scm_set_cdr1:
print_str("error: non pair argument to 'set-cdr': ")
	moveq	#-1,d0
	jbsr	scm_write
print_str("\012")
	jmp	dump
scm_set_cdr2:
	jmp	error_4
end_primitive

define_primitive(null)
	bpls	scm_null2
	addql	#1,d1
	beqs	scm_null1
	moveq	#Cst_false,d1
	rts
scm_null1:
	moveq	#Cst_true,d1
	rts
scm_null2:
	jmp	error_4
end_primitive

define_primitive(num_eq)
	bnes	scm_num_eq3
	btst	d1,d5		| is arg1 a fixnum?
	beqs	scm_num_eq2
	btst	d2,d5		| is arg2 a fixnum?
	beqs	scm_num_eq2
	cmpl	d2,d1
	beqs	scm_num_eq1
	moveq	#Cst_false,d1
	rts
scm_num_eq1:
	moveq	#Cst_true,d1
	rts
scm_num_eq2:
print_str("error: non fixnum argument(s) to '='\012")
	jmp	dump
scm_num_eq3:
	jmp	error_4
end_primitive

define_primitive(num_lt)
	bnes	scm_num_lt3
	btst	d1,d5		| is arg1 a fixnum?
	beqs	scm_num_lt2
	btst	d2,d5		| is arg2 a fixnum?
	beqs	scm_num_lt2
	cmpl	d2,d1
	blts	scm_num_lt1
	moveq	#Cst_false,d1
	rts
scm_num_lt1:
	moveq	#Cst_true,d1
	rts
scm_num_lt2:
print_str("error: non fixnum argument(s) to '<'\012")
	jmp	dump
scm_num_lt3:
	jmp	error_4
end_primitive

define_primitive(num_gt)
	bnes	scm_num_gt3
	btst	d1,d5		| is arg1 a fixnum?
	beqs	scm_num_gt2
	btst	d2,d5		| is arg2 a fixnum?
	beqs	scm_num_gt2
	cmpl	d2,d1
	bgts	scm_num_gt1
	moveq	#Cst_false,d1
	rts
scm_num_gt1:
	moveq	#Cst_true,d1
	rts
scm_num_gt2:
print_str("error: non fixnum argument(s) to '>'\012")
	jmp	dump
scm_num_gt3:
	jmp	error_4
end_primitive

define_primitive(plus)
	bnes	scm_plus2
	btst	d1,d5		| is arg1 a fixnum?
	beqs	scm_plus5
	btst	d2,d5		| is arg2 a fixnum?
	beqs	scm_plus5
	addl	d2,d1
	bvss	scm_plus1
	rts
scm_plus1:
print_str("error: overflow on '+'\012")	| for now signal an error, eventually
	jmp	dump			| should convert result to bignum
scm_plus2:
	bpls	scm_plus3
	btst	d1,d5		| is arg1 a fixnum?
	beqs	scm_plus5
	rts
scm_plus3:
	subql	#1,d0
	bnes	scm_plus4
	moveq	#0,d0
	rts
scm_plus4:
	addql	#1,d0
	jmp	error_4		| for now signal an error (should be n-ary)
scm_plus5:
print_str("error: non fixnum argument(s) to '+'\012")
	jmp	dump
end_primitive

define_primitive(minus)
	bnes	scm_minus2
	btst	d1,d5		| is arg1 a fixnum?
	beqs	scm_minus4
	btst	d2,d5		| is arg2 a fixnum?
	beqs	scm_minus4
	subl	d2,d1
	bvss	scm_minus1
	rts
scm_minus1:
print_str("error: overflow on '-'\012")	| for now signal an error, eventually
	jmp	dump			| should convert result to bignum
scm_minus2:
	bpls	scm_minus3
	btst	d1,d5		| is arg1 a fixnum?
	beqs	scm_minus4
	negl	d1
	bvss	scm_minus1
	rts
scm_minus3:
	jmp	error_4		| for now signal an error (should be n-ary)
scm_minus4:
print_str("error: non fixnum argument(s) to '-'\012")
	jmp	dump
end_primitive

define_primitive(times)
	bnes	scm_times2
	btst	d1,d5		| is arg1 a fixnum?
	beqs	scm_times5
	btst	d2,d5		| is arg2 a fixnum?
	beqs	scm_times5
	asrl	#3,d1
	muls	d2,d1
	bvss	scm_times1
	rts
scm_times1:
print_str("error: overflow on '*'\012")	| for now signal an error, eventually
	jmp	dump			| should convert result to bignum
scm_times2:
	bpls	scm_times3
	btst	d1,d5		| is arg1 a fixnum?
	beqs	scm_times5
	rts
scm_times3:
	subql	#1,d0
	bnes	scm_times4
	moveq	#8,d0
	rts
scm_times4:
	addql	#1,d0
	jmp	error_4		| for now signal an error (should be n-ary)
scm_times5:
print_str("error: non fixnum argument(s) to '*'\012")
	jmp	dump
end_primitive

define_primitive(quotient)
	bnes	scm_quotient1
	btst	d1,d5		| is arg1 a fixnum?
	beqs	scm_quotient2
	btst	d2,d5		| is arg2 a fixnum?
	beqs	scm_quotient2
	divs	d2,d1
	extl	d1
	asll	#3,d1
	rts
scm_quotient1:
	jmp	error_4		| for now signal an error (should be n-ary)
scm_quotient2:
print_str("error: non fixnum argument(s) to '/'\012")
	jmp	dump
end_primitive

define_primitive(minus_1)
	bpls	scm_minus_11
	btst	d1,d5		| is arg1 a fixnum?
	beqs	scm_minus_12
	subql	#8,d1
	rts
scm_minus_11:
	jmp	error_4
scm_minus_12:
print_str("error: non fixnum argument to '-1+'\012")
	jmp	dump
end_primitive

define_primitive(force)
	bpls	scm_force2
	btst	d1,d7
	beqs	scm_force1
	movl	d1,a0
	tstw	a0@(-2)
	bpls	scm_force1
	moveq	#1,d0
	jmp	a0@
scm_force1:
	jmp	a6@(-510)
scm_force2:
	jmp	error_4
end_primitive

define_primitive(write)
	bpls	scm_write1
	btst	d1,d5	| is it a fixnum?
	beqs	not_fixnum

	movl	d1,d2
	asrl	#3,d2
print_int("%d",d2)
	moveq	#Cst_undefined,d1
	rts

scm_write1:
	jmp	error_4

not_fixnum:
	btst	#0,d1
	beqs	not_char2
	swap	d1
	cmpw	#-2,d1
	bnes	not_char1

	swap	d1
	asrw	#1,d1

print_chr("#\134%c",d1)
	moveq	#Cst_undefined,d1
	rts

not_char1:
	swap	d1
not_char2:
	btst	d1,d6	| is it a pair?
	beq	not_pair

print_str("\050")

next_pair:
	movl	d1,a0
	movl	a0@,d1
	movl	a0@-,a7@-
	moveq	#-1,d0
	jbsr	scm_write

	movl	a7@+,d1
	cmpl	#Cst_nil,d1
	bnes	not_end

print_str("\051")
	moveq	#Cst_undefined,d1
	rts

not_end:
	btst	d1,d6	| is it a pair?
	bnes	pair_again

print_str(" . ")

	moveq	#-1,d0
	jbsr	scm_write

print_str("\051")
	moveq	#Cst_undefined,d1
	rts

pair_again:
print_str(" ")
	bra	next_pair

not_pair:
	cmpl	#Cst_nil,d1
	bnes	not_nil

print_str("\050\051")
	moveq	#Cst_undefined,d1
	rts

not_nil:
	cmpl	#Cst_false,d1
	bnes	not_false

print_str("#f")
	moveq	#Cst_undefined,d1
	rts

not_false:
	cmpl	#Cst_true,d1
	bnes	not_true

print_str("#t")
	moveq	#Cst_undefined,d1
	rts

not_true:
	cmpl	#Cst_undefined,d1
	bnes	not_undef

print_str("#<undefined>")
	moveq	#Cst_undefined,d1
	rts

not_undef:
	btst	d1,d7
	beq	unknown

	movl	d1,a0
	movw	a0@(-2),d0

	bpl	other_hunks

	cmpw	#Closure_type,d0
	bhi	write_closure

	cmpw	#Primitive_type,d0
	bhi	write_primitive

print_str("#<return address into ")

	subw	#Sub_proc_type,d0
	extl	d0
	asll	#2,d0

	subl	d0,a0
	movl	a0,d1
	moveq	#-1,d0
	jbsr	scm_write

print_str(">")

	moveq	#Cst_undefined,d1
	rts

write_primitive:

print_str("#<procedure")

	subw	#Primitive_type,d0
	extl	d0
	asll	#2,d0

	addl	d0,a0
	subql	#6,a0
	movl	a0@,d1
	cmpl	#Cst_nil,d1
	beqs	write_primitive1

print_str(" ")

	moveq	#-1,d0
	jbsr	scm_write

write_primitive1:

print_str(">")

	moveq	#Cst_undefined,d1
	rts

write_closure:

	addql	#2,a0
	movl	a0@,d1
	moveq	#-1,d0
	jbsr	scm_write

	moveq	#Cst_undefined,d1
	rts

other_hunks:
	asrw	#8,d0

	cmpb	#Vector_type,d0
	bne	other_1

print_str("#\050")

	subql	#2,a0
	movl	a0@+,d0
	andl	#0xFFFFFF,d0
	beqs	vector_end

write_vector_loop:

	movl	a0@+,d1
	movl	d0,a7@-
	movl	a0,a7@-
	moveq	#-1,d0
	jbsr	scm_write
	movl	a7@+,a0
	movl	a7@+,d0

	subql	#1,d0
	beqs	vector_end

print_str(" ")
	bra	write_vector_loop

vector_end:
print_str("\051")
	moveq	#Cst_undefined,d1
	rts

other_1:
	cmpb	#String_type,d0
	bne	other_2

print_str("\042")

	subql	#2,a0
	movl	a0@+,d0
	andl	#0xFFFFFF,d0
	beqs	string_end

write_string_loop:

	movb	a0@+,d1

	cmpb	#34,d1
	beqs	write_escape
	cmpb	#92,d1
	bnes	write_string_char

write_escape:
print_chr("%c",#92)

write_string_char:
print_chr("%c",d1)

	subql	#1,d0
	bnes	write_string_loop

string_end:
print_str("\042")
	moveq	#Cst_undefined,d1
	rts

other_2:
	cmpb	#Symbol_type,d0
	bne	other_3

	movl	a0@(2),a0

	subql	#2,a0
	movl	a0@+,d0
	andl	#0xFFFFFF,d0
	beqs	symbol_end

write_symbol_loop:

	movb	a0@+,d1

print_chr("%c",d1)

	subql	#1,d0
	bnes	write_symbol_loop

symbol_end:
	moveq	#Cst_undefined,d1
	rts

other_3:

unknown:
print_str("#<unknown>")
	moveq	#Cst_undefined,d1
	rts
end_primitive

define_primitive(newline)
	subqw	#1,d0
	bne	scm_newline1
print_str("\012")
	moveq	#Cst_undefined,d1
	rts
scm_newline1:
	addqw	#1,d0
	jmp	error_4
end_primitive

define_primitive(assq)
	bnes	scm_assq4
scm_assq1:
	btst	d2,d6
	beqs	scm_assq2
	movl	d2,a0
	movl	a0@,d3
	movl	a0@-,d2
	btst	d3,d6
	beqs	scm_assq3
	movl	d3,a1
	cmpl	a1@,d1		| use eq? for now (should use eqv?)
	bnes	scm_assq1
	movl	d3,d1
	rts	
scm_assq2:
	moveq	#Cst_false,d1
	rts	
scm_assq3:
print_str("error: illformed a-list to 'assq'\012")
	jmp	dump
scm_assq4:
	jmp	error_4
end_primitive

define_primitive(symbol)
	bpls	scm_symbol2
	btst	d1,d7
	beqs	scm_symbol1
	movl	d1,a0
	cmpb	#Symbol_type,a0@(-2)
	bnes	scm_symbol1
	moveq	#Cst_true,d1
	rts
scm_symbol1:
	moveq	#Cst_false,d1
	rts
scm_symbol2:
	jmp	error_4
end_primitive

define_primitive(vector)
	bpls	scm_vector2
	btst	d1,d7
	beqs	scm_vector1
	movl	d1,a0
	cmpb	#Vector_type,a0@(-2)
	bnes	scm_vector1
	moveq	#Cst_true,d1
	rts
scm_vector1:
	moveq	#Cst_false,d1
	rts
scm_vector2:
	jmp	error_4
end_primitive

define_primitive(string)
	bpls	scm_string2
	btst	d1,d7
	beqs	scm_string1
	movl	d1,a0
	cmpb	#String_type,a0@(-2)
	bnes	scm_string1
	moveq	#Cst_true,d1
	rts
scm_string1:
	moveq	#Cst_false,d1
	rts
scm_string2:
	jmp	error_4
end_primitive

define_primitive(procedure)
	bpls	scm_procedure2
	btst	d1,d7
	beqs	scm_procedure1
	movl	d1,a0
	tstw	a0@(-2)
	bpls	scm_procedure1
	moveq	#Cst_true,d1
	rts
scm_procedure1:
	moveq	#Cst_false,d1
	rts
scm_procedure2:
	jmp	error_4
end_primitive

define_primitive(number)
	bpls	scm_number2
	btst	d1,d5
	beqs	scm_number1
	moveq	#Cst_true,d1
	rts
scm_number1:
	moveq	#Cst_false,d1
	rts
scm_number2:
	jmp	error_4
end_primitive

define_primitive(char)
	bpls	scm_char2
	btst	#0,d1
	beqs	scm_char1
	swap	d1
	cmpw	#-2,d1
	bnes	scm_char1
	moveq	#Cst_true,d1
	rts
scm_char1:
	moveq	#Cst_false,d1
	rts
scm_char2:
	jmp	error_4
end_primitive

| -----------------------------------------------------------------------------
|
| Run-time procedures (written in Scheme):


procedure_object_begin(eval_print)

	pea	eval_print1
	moveq	#1,d0
	jmp	entry		| enter program
sub_procedure(eval_print1)

print_str("Result: ")

	pea	eval_print2
	moveq	#-1,d0
	jmp	scm_write
sub_procedure(eval_print2)

print_str("\012")

	jmp	exit
procedure_object_constants
	.long	Cst_nil
procedure_object_end

procedure_object_begin(dump)

print_str("\012Stack dump:\012")

dump1:
	cmpl	#stack_top,a7
	bccs	dump2

print_str("> ")
	movl	sp@+,d1
	moveq	#-1,d0
	jbsr	scm_write

print_str("\012")

	bras	dump1

dump2:
	jmp	exit
procedure_object_constants
procedure_object_end

define_primitive(make_promise)

| (define (make-promise thunk)
|   (let ((computed #f))
|     (lambda ()
|       (if computed thunk
|         (begin
|           (set! thunk (thunk))
|           (set! computed #t)
|           thunk)))))

	bpls	scm_make_promise2
	cmpl	a6@,a7
	bhis	scm_make_promise3
scm_make_promise2:
	jmp	a6@(-522)
scm_make_promise3:
	movl	d1,a7@-
	movl	a7@,d0
	movl	d0,a4@-
	movl	a4,a7@-
	movl	d0,a4@-
	moveq	#Cst_false,d0
	movl	d0,a7@-
	movl	a7@,d0
	movl	d0,a4@-
	movl	a4,a7@-
	movl	d0,a4@-
	movw	#-8189,a5@+
	movl	a5,d1
	movw	#20153,a5@+
	lea	scm_make_promise4,a0
	movl	a0,a5@+
	movl	a7@(8),a5@+
	movl	a7@,a5@+
	addw	#16,a7
	rts	
sub_procedure(scm_make_promise4)
	subql	#1,d0
	bnes	scm_make_promise5
	cmpl	a6@,a7
	bhis	scm_make_promise6
scm_make_promise5:
	jmp	a6@(-516)
scm_make_promise6:
	subql	#6,a7@
	movl	a7@,a0
	movl	a0@(10),a0
	movl	a0@,d1
	addql	#3,d1
	bcs	scm_make_promise7
	movl	a7@,a0
	movl	a0@(6),a0
	movl	a0@,d1
	addql	#4,a7
	rts	
scm_make_promise7:
	movl	a7@,a0
	movl	a0@(6),a0
	movl	a0@,d0
	pea	scm_make_promise9
	btst	d0,d7
	beqs	scm_make_promise8
	movl	d0,a0
	tstw	a0@(-2)
	bpls	scm_make_promise8
	moveq	#1,d0
	jmp	a0@
scm_make_promise8:
	jmp	a6@(-510)
sub_procedure(scm_make_promise9)
	movl	a7@,a0
	movl	a0@(6),a0
	movl	d1,a0@
	moveq	#Cst_true,d1
	movl	a7@,a0
	movl	a0@(10),a0
	movl	d1,a0@
	movl	a7@,a0
	movl	a0@(6),a0
	movl	a0@,d1
	addql	#4,a7
	rts	
end_primitive

define_primitive(memv)

| (define (memv obj list)
|  (if (pair? list)
|    (if (eqv? (car list) obj)
|      list
|      (memv obj (cdr list)))
|    #f))

	bnes	scm_memv1
	cmpl	a6@,a7
	bhis	scm_memv3
scm_memv1:
	jmp	a6@(-522)
scm_memv2:
	movl	a0@-,d2
scm_memv3:
	btst	d2,d6
	beqs	scm_memv4
	movl	d2,a0
	cmpl	a0@,d1		| use eq? for now (should use eqv?)
	bnes	scm_memv2
	movl	d2,d1
	rts	
scm_memv4:
	moveq	#Cst_false,d1
	rts	
end_primitive

define_primitive(list)

| (define (list . l) l)

	movw	#1,a0
	lea	scm_list1,a1
	jmp	a6@(-534)
sub_procedure(scm_list1)
	movl	a7@,d1
	addql	#4,a7
	rts	
end_primitive

define_primitive(vect)

| (define (vector . l) (list->vector l))

	movw	#1,a0
	lea	scm_vect1,a1
	jmp	a6@(-534)
sub_procedure(scm_vect1)
	movl	a7@+,d1
	moveq	#-1,d0
	jmp	scm_list_vector
end_primitive

| -----------------------------------------------------------------------------
|
| The program:

